#=====================================================================
#   HTTP.pm : Simple HTTP GET/POST/HEAD method client.
#
#   Copyright(c) Nobuchika Oishi (BSC CONSULTING).
#
#   e-mail : bigstone@my.email.ne.jp
#   support: http://www.din.or.jp/~bigstone/cgilab/index.html
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND
#   ANY EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED.
#
#   ver 1.2.3 : updated last on 2002/05/16
#
#=====================================================================
package HTTP;

use Symbol;
use Carp;
use TCP;

use strict;
use vars qw(
    $DYNAMIC_FILE_UPLOAD $CHECK_ALLOW_HEADERS $UNDEF_ERROR_DOCUMENT
    %ALLOW_HEADERS $GUNZIP_TEMP_DIR $GUNZIP_COMMAND $MAX_SOCKET_READ
    $VERSION $CRLF $CRLF_ANY
);

#---------------------------------------------------------------------
$DYNAMIC_FILE_UPLOAD  = 1;  # set 0 or 1
$CHECK_ALLOW_HEADERS  = 1;  # set 0 or 1
$UNDEF_ERROR_DOCUMENT = 1;  # set 0 or 1

$GUNZIP_TEMP_DIR = '/tmp';
$GUNZIP_COMMAND  = '/usr/bin/gunzip -cq';

$MAX_SOCKET_READ = 16384;   # set number of less than 32766 (bytes)
#---------------------------------------------------------------------

$VERSION  = '1.2.3';

$CRLF_ANY = "\015?\012";
$CRLF     = "\015\012";

%ALLOW_HEADERS = (
    date                => 1,
    expires             => 1,
    if_modified_since   => 1,
    if_unmodified_since => 1,
    last_modified       => 1,
    client_date         => 1,
    title               => 1,
    content_encoding    => 1,
    content_language    => 1,
    connection          => 1,
    server              => 1,
    from                => 1,
    if_match            => 1,
    if_none_match       => 1,
    accept              => 1,
    accept_charset      => 1,
    accept_encoding     => 1,
    accept_language     => 1,
    referer             => 1,
    user_agent          => 1,
    mime_version        => 1,
    range               => 1,
    cookie              => 1,
    warning             => 1,
    www_authenticate    => 1,
    authorization       => 1,
    proxy_authenticate  => 1,
    proxy_authorization => 1,
);

sub new {
    my $class = shift;
    my $self = {
        request => {},
        header  => {},
        status  => [],
        error   => 0,
        debug   => 0,
    };
    bless $self, ref $class || $class;
    return @_ ? $self->connect(@_) : $self;
}

sub connect {
    my $self = shift;
    my($host, $port, $proxy, $time, $k, $v);
    while (($k, $v) = splice(@_, 0, 2)){
        next if $v eq "";
        $k = lc($k);
        if($k eq 'host'){
            ($host, $port) = $v =~ m!^([^:/]+):?(\d*)$!o;
        } elsif($k eq 'proxy'){
            $proxy = $v;
        } elsif($k eq 'port'){
            $port = $v;
        } elsif($k eq 'timeout'){
            $time = $v;
        }
    }
    return undef unless $host;
    $self = $self->new() unless ref $self;
    $port ||= 80;
    my($peer_host, $peer_port);
    if($proxy){
        if($proxy =~ /^([^:]+):(\d+)$/o){
            $peer_host = $1;
            $peer_port = $2;
        } else {
            $peer_host = $proxy;
            $peer_port = 80;
        }
        $self->{proxy} = $proxy;
    } else {
        $peer_host = $host;
        $peer_port = $port;
    }
    $self->{socket} = TCP->connect($peer_host, $peer_port, $time) || return undef;
    $self->{host} = $host;
    $self->{port} = $port;
    return $self;
}

sub close {
    my $self = shift;
    if(my $sock = delete $self->{socket}){
        $sock->close();
    }
}

sub request {
    my $self = shift;
    my(%parm, $method, $host, $port, $proxy, $timeout, $name, $value, $k, $v);
    while (($name, $value) = splice(@_, 0, 2)){
        next if $value eq "";
        $name = lc($name);
        if($name =~ /^(get|head|post)$/o){
            $method = uc($name);
            while (($k, $v) = splice(@{$value}, 0, 2)){
                $k = lc($k);
                if($k eq 'host'){
                    $host = $v;
                } elsif($k eq 'port'){
                    $port = $v;
                } elsif($k eq 'uri'){
                    $parm{uri} = $v;
                }
            }
        } elsif($name eq 'proxy'){
            $proxy = $value;
        } elsif($name eq 'timeout'){
            $timeout = $value;
        } else {
            $parm{$name} = $value;
        }
    }
    if(defined $parm{'host'}){
        $host = delete $parm{'host'};
    }
    if(defined $parm{'port'}){
        $port = delete $parm{'port'};
    }
    if(defined $parm{'method'}){
        $method = delete $parm{'method'};
        $method = uc($method);
    }
    return undef unless($method && $host);
    $self = $self->new(
        host    => $host,
        port    => $port || 80,
        proxy   => $proxy,
        timeout => $timeout,
    ) || return undef;
    if($v = delete $parm{save}){
        $self->savefile($v);
    }
    $self->{atonce} = 1;
    $self->_exec($method, \%parm);
    if($method eq 'HEAD'){
        return wantarray ? ($self, %{$self->{header}}) : $self;
    } else {
        return wantarray ? ($self, $self->{content}) : $self->{content};
    }
}

sub GET  { shift->_request('GET',  @_); }
sub HEAD { shift->_request('HEAD', @_); }
sub POST { shift->_request('POST', @_); }

sub _request {
    my $self   = shift;
    my $method = shift;
    return undef unless ref $self;
    my(%parm, $k, $v);
    while (($k, $v) = splice(@_, 0, 2)){
        $parm{lc($k)} = $v if $v ne "";
    }
    $self->{header}  = {};
    $self->{content} = "";
    $self->{status}  = [];
    $self->{error}   = 0;
    $self->{atonce}  = 0;
    if($v = delete $parm{save}){
        $self->savefile($v);
    } else {
        delete $self->{save};
        delete $self->{write};
    }
    $self->_exec($method, \%parm);
    if($method eq 'HEAD'){
        return wantarray ? %{$self->{header}} : $self->{header};
    } else {
        return wantarray ? ($self->{content}, %{$self->{header}}) : $self->{content};
    }
}

sub _exec {
    my $self = shift;
    my $sock = $self->{socket};
    unless($sock){
        $self->_set_status(900, "No Connection"); $self->{logs} = [];
        return undef;
    }
    my($method, $uri, $header, $content) = $self->_make_request(@_);
    if($header){
        if($method eq 'POST' && !ref($content)){
            $header .= join('', 'Content-Length: ', length($content), $CRLF);
        }
        $header .= $CRLF;
       HTTPEXEC:
        {
            $self->_send_string($sock, $header) || last HTTPEXEC;
            if(ref($content)){
                while (my $str = &$content()){
                    $self->_send_string($sock, $str) || last HTTPEXEC;
                }
            } else {
                $self->_send_string($sock, $content) || last HTTPEXEC;
            }
            $self->_response($sock, $method, $uri);
        }
    }
    if($self->{error} || $self->{atonce} || $self->{header}->{connection} !~ /^Keep-Alive/oi){
        $self->close();
    }
    if($UNDEF_ERROR_DOCUMENT){
        $self->{content} = undef unless $self->is_success();
    }
}

sub _make_request {
    my $self = shift;
    my($method, $parm) = @_;
    my($header, $content_header, $content, $k, $v);
    my $host = $self->{host};
    my $port = $self->{port} != 80 ? join('', ':', $self->{port}) : "";
    if(defined $parm->{header}){
        while (($k, $v) = splice(@{$parm->{header}}, 0, 2)){
            if($CHECK_ALLOW_HEADERS){
                unless(_is_allow_header($k)){
                    $self->_set_status(901, "Not Allow Header: $k");
                    return undef;
                }
            }
            $self->{request}->{$k} = $v;
        }
    }
    if($method eq 'POST'){
        if($parm->{enctype} eq 'form-data'){
            ($content_header, $content) = $self->_make_content_form_data($parm);
        } else {
            ($content_header, $content) = $self->_make_content_urlencoded($parm);
        }
    }
    if($parm->{version} =~ /^[\d.]+$/o){
        $self->{http} = $parm->{version};
    }
    my $uri  = $parm->{uri}  || '/';
    my $http = $self->{http} || '1.1';
    if($http eq '1.0'){
        $self->{request}->{Connection} = 'close'; delete $self->{request}->{connection};
    } else {
        $self->{request}->{Connection} ||= (delete $self->{request}->{connection} || 'close');
    }
    $self->{request}->{Accept} ||= (delete $self->{request}->{accept} || '*/*');
    if($self->{proxy}){
        $uri = join('', 'http://', $host, $port, $uri);
    }
    $header  = "$method $uri HTTP/$http$CRLF";
    $header .= "Host: $host$port$CRLF";
    while (($k, $v) = each %{$self->{request}}){
        $header .= "$k: $v$CRLF";
    }
    $header .= $content_header if $content_header;
    if(exists $parm->{debug}){
        $self->set_debug($parm->{debug});
    }
    $self->{logs} = [] if $self->{debug};   # clear logs every request times.
    return ($method, $uri, $header, $content);
}

sub _make_content_urlencoded {
    my($self, $parm) = @_;
    my($header, $content, @query, $k, $v);
    if(exists $parm->{content}){
        if(!ref($parm->{content})){
            $content = $parm->{content};  # must be url-encoded!
        } else {
            while (($k, $v) = splice(@{$parm->{content}}, 0, 2)){
                next if ref $v;
                push(@query, join('=', encode_url($k), encode_url($v)));
            }
            $content = join('&', @query) if @query;
        }
    }
    $content .= $CRLF;
    $header   = "Content-Type: application/x-www-form-urlencoded$CRLF";
    return ($header, $content);
}

sub _make_content_form_data {
    my($self, $parm) = @_;
   #---------------------
    require MTYPE;
   #---------------------
    my($header, $content, $part, @parts, $k, $v);
    my($filename, $guess_encode, $file, $type, $encode);
    my $dynamic = 0;
    if(!exists $parm->{content} || !ref($parm->{content})){
        $header  = "Content-Type: application/x-www-form-urlencoded$CRLF";
        $content = $parm->{content} . $CRLF;
        retrun ($header, $content);
    } else {
        while (($k, $v) = splice(@{$parm->{content}}, 0, 2)){
            $k =~ s/([\\\"])/\\$1/go;
            if(!ref($v)){
                push(@parts, qq(Content-Disposition: form-data; name="$k"$CRLF$CRLF$v));
                next;
            } else {
                ($file, $type, $encode) = @{$v};
                next unless $file;
                ($filename) = $file =~ m!([^/:]+)$!o;
                unless($type){
                    ($type, $guess_encode) = MTYPE::guess_type($file);
                }
                $encode ||= $guess_encode;
                unless($encode){
                    $encode = 'binary' if($type =~ m!^image/!o);
                }
                $part  = qq(Content-Disposition: form-data; name="$k"; filename="$filename"$CRLF);
                $part .= qq(Content-Type: $type$CRLF);
                $part .= qq(Content-Transfer-Encoding: $encode$CRLF) if $encode;
                $part .= $CRLF;
                my $fh = Symbol::gensym();
                open($fh, "< $file") or Carp::croak("Can't open file: $!");
                binmode($fh);
                if($DYNAMIC_FILE_UPLOAD){
                    $file = $fh;
                } else {
                    local($/) = undef;
                    $file = <$fh>;
                    CORE::close($fh);
                }
                if(ref($file)){
                    push(@parts, [$part, $file]); $dynamic++;
                } else {
                    push(@parts, join('', $part, $file));
                }
            }
        }
    }
    my $boundary = join('_', '------=', $self->_boundary(10));
    $header = qq(Content-Type: multipart/form-data; boundary="$boundary"$CRLF);
    if($dynamic){
        for(1..@parts-1){
            splice(@parts, $_*2-1, 0, "$CRLF--$boundary$CRLF");
        }
        unshift(@parts, "--$boundary$CRLF");
        push(@parts, "$CRLF--$boundary--$CRLF");
        my $fsize  = 0;
        my $length = 0;
        foreach (@parts){
            if(ref $_){
                ($part, $file) = @$_;
                unless(-f $file && ($fsize = -s _)){
                    undef($length);
                    last;
                } else {
                    $length += $fsize + length($part);
                }
            } else {
                $length += length;
            }
        }
        if(defined $length){
            $header .= "Content-Length: $length$CRLF";
        }
        $content = sub {
            for(;;){
                unless(@parts){
                    if(defined $length && $length != 0){
                        Carp::croak("Length of data sent did not match.");
                    }
                    return undef;
                }
                my $part = shift(@parts);
                unless(ref $part){
                    while (@parts && !ref($parts[0])){
                        $part .= shift(@parts);
                    }
                    $length -= length($part) if(defined $length);
                    return $part;
                }
                my($buff, $fh) = @{$part};
                my $offset = length($buff);
                my $getlen = read($fh, $buff, 2048, $offset);
                if($getlen){
                    $offset += $getlen;
                    unshift(@parts, ['', $fh]);
                } else {
                    CORE::close($fh);
                }
                if($offset){
                    defined $length && ($length -= $offset);
                    return $buff;
                }
            }
        };
    } else {
        $content = "--$boundary$CRLF". join("$CRLF--$boundary$CRLF", @parts) ."$CRLF--$boundary--$CRLF";
    }
    return ($header, $content);
}

sub _send_string {
    my($self, $sock, $str) = @_;
    if($self->{debug}){
        push(@{$self->{logs}}, $str);
    }
    unless($sock->write($str, '1')){
        $self->_set_status(902, "Socket I/O Error");
        return undef;
    }
    1;
}

sub _response {
    my($self, $sock, $method, $uri) = @_;
    my $response = $sock->read();
    if($response =~ m!^HTTP/\d\.\d +(\d+) +([^\r\n]*)!oi){
        $self->_set_status($1, $2);
    } else {
        $self->_set_status(902, "Socket I/O Error");
        return undef;
    }
    $response = $sock->read_until("$CRLF_ANY$CRLF_ANY");
    $response =~ s/$CRLF_ANY/\n/sgo;
    $response =~ s/\n / /sgo;
    $response =~ s/\n+$//so;
    my($k, $v);
    foreach (split(/\n/, $response)){
        ($k, $v) = /^ *([^:]+) *: *(.*)/o;
        $self->{header}->{lc($k)} = $v;
    }
    if($method eq 'HEAD' || !defined $self->{header}->{'content-type'}){
        return 1;
    }
    my($type) = $self->{header}->{'content-type'} =~ m!^(text|image)/!oi;
    my $gzip  = $self->{header}->{'content-encoding'} =~ /gzip/oi;
    my $fh = undef;
    $type = lc($type); $response = "";
    if($v = delete $self->{save}){
        if($v =~ s!/\*$!/!o){
            if($uri =~ m!([^/]+)$!o){
                $v .= $1;
            } else {
                $self->_set_status(904, "Invalid Save parameters");
                return undef;
            }
        }
        if(-e $v){
            if(lc($self->{write}) eq 'norep'){
                $self->_set_status(905, "File already exist: $v");
                return undef;
            }
        }
        $fh = Symbol::gensym();
        unless(open($fh, "> $v")){
            $self->_set_status(906, "File can not open: $v");
            return undef;
        }
        binmode($fh) if $type ne 'text';
    } else {
        unless($type){
            $self->_set_status(903, "Binary File Request");
            return undef;
        }
    }
    if($self->{header}->{'transfer-encoding'} =~ /chunked/oi){
        while (1){
            $v = $sock->read();
            $v =~ s/[\r\n]+$//so;
            $v = hex("0x$v");
            last unless $v;
            $response .= _content($sock, $v, $type, $fh);
            $sock->read();
        }
        while ($v = $sock->read()){
            last if($v =~ /^$CRLF_ANY$/so);
        }
    } else {
        if(($v = $self->{header}->{'content-length'}) ne ""){
            $response = _content($sock, $v, $type, $fh) if $v > 0;
        } else {
            $response = _content($sock, $v, $type, $fh);
        }
    }
    if($response){
        if($gzip && $GUNZIP_TEMP_DIR && $GUNZIP_COMMAND){
            unless(_gunzip(\$response)){
                $self->_set_status(907, "Can't extract [gzip] response");
                return undef;
            }
        }
        if($type eq 'text'){
            $response =~ s/$CRLF_ANY/\n/sgo;
        }
        print $fh $response if $fh;
    }
    if($fh){
        CORE::close($fh); $response = "";
    }
    $self->{content} = $response;
    1;
}

sub _content {
    my($sock, $size, $type, $fh) = @_;
    my $res = "";
    if($type){
        $res = $size > 0 ? $sock->read($size) : $sock->read_until();
    } elsif($fh){
        if($size > 0){
            my($buff, $len);
            while ($size){
                $len  = $size >= $MAX_SOCKET_READ ? $MAX_SOCKET_READ : $size;
                $buff = $sock->read($len) || last;
                print $fh $buff;
                $size -= $len;
            }
        } else {
            print $fh $sock->read_until();
        }
    }
    $res;
}

sub _gunzip {
    my $rres = shift;
    my($temp, $success);
    my $fh = Symbol::gensym();
    my $n  = 0;
   TEMP_WRITE: {
        if($temp){
            return undef if(++$n > 9);
            sleep(1);
        }
        if($^O =~ /^(MS)?Win/oi){
            $temp = sprintf("%04d%04d.%03d", $$, int(rand(9999)), $n);
        } else {
            $temp = join('.', time, $$, $<, int(rand(9999)), $n, 'gz');
        }
        if(-e "$GUNZIP_TEMP_DIR/$temp"){
            if(-M _ < .0035){
                redo TEMP_WRITE;
            }
        }
        open($fh, "> $GUNZIP_TEMP_DIR/$temp") or redo TEMP_WRITE;
        binmode($fh);
        print $fh $$rres;
        CORE::close($fh);
    }
    if(chdir($GUNZIP_TEMP_DIR)){
        eval {
            $$rres = `$GUNZIP_COMMAND $temp`;
        };
        unless($@){
            $success = 1 if $$rres ne "";
        }
        $0 =~ m!^(.+?)[/:\\][^/:\\]+$!o && chdir $1;
    }
    unlink "$GUNZIP_TEMP_DIR/$temp";
    return $success;
}

sub _set_status {
    my($self, $status, $reason) = @_;
    @{$self->{status}} = ($status, $reason);
    if($status >= 900){
        $self->{error} = 1;
    }
    return undef;
}

sub header {
    my($self, $name) = @_;
    if(defined $name){
        return $self->{header}->{lc($name)};
    }
    return wantarray ? %{$self->{header}} : $self->{header};
}

sub content {
    shift->{content};
}

sub status {
    my $self = shift;
    return wantarray ? @{$self->{status}} : $self->{status}->[0];
}

sub http_version {
    my($self, $ver) = @_;
    $self->{http} = $ver if($ver =~ /^[\d.]+$/o);
}

sub is_success {
    my $code = @_ > 1 ? $_[1] : shift->{status}->[0];
    $code >= 200 && $code < 300;
}

sub is_keep_alive {
    my $self = shift;
    defined $self->{socket} && $self->{header}->{connection} =~ /^Keep-Alive/oi;
}

sub savefile {
    my $self = shift;
    ($self->{save}, $self->{write}) = ref $_[0] ? @{ $_[0] } : @_;
}

sub _boundary {
    my $self = shift;
    my $size = shift || 10;
    my $b = encode_base64(join('', map chr(rand(256)), 1..$size*3), "");
    $b =~ s/[\W]/X/go;
    $b;
}

sub _is_allow_header {
    my($name) = @_;
    return undef unless $name;
    $name =~ tr/-/_/;
    return undef unless exists $ALLOW_HEADERS{lc($name)};
    1;
}

sub _header {
    my $self = shift;
    my($header, $value) = @_;
    return undef unless defined $value;
    $self->{request}->{$header} = $value;
}

sub _gmdate {
    my $self = shift;
    my($header, $time) = @_;
   #---------------------
    require DATE;
   #---------------------
    my $date = $self->{request}->{$header};
    if(defined $time){
        $self->{request}->{$header} = DATE::time2str($time);
    }
    return $date ? DATE::str2time($date) : undef;
}

sub _basic_auth {
    my $self = shift;
    my($header, $name, $pass) = @_;
    return undef unless defined $name;
    return undef if($name =~ /:/o);
    $pass = '' unless defined $pass;
    $self->{request}->{$header} = join(' ', 'BASIC', encode_base64("$name:$pass", ""));
}

sub encode_url {
    my $str = shift;
    $str =~ s/([^\-_.!*'()0-9A-Za-z])/sprintf("%%%02X", ord($1))/ego;
    $str;
}

#---------------------------------------------
#   Functions for request header supplement
#---------------------------------------------
sub date                { shift->_gmdate('Date',                @_); }
sub expires             { shift->_gmdate('Expires',             @_); }
sub if_modified_since   { shift->_gmdate('If-Modified-Since',   @_); }
sub if_unmodified_since { shift->_gmdate('If-Unmodified-Since', @_); }
sub last_modified       { shift->_gmdate('Last-Modified',       @_); }
sub client_date         { shift->_gmdate('Client-Date',         @_); }
sub title               { shift->_header('Title',               @_); }
sub content_encoding    { shift->_header('Content-Encoding',    @_); }
sub content_language    { shift->_header('Content-Language',    @_); }
sub connection          { shift->_header('Connection',          @_); }
sub server              { shift->_header('Server',              @_); }
sub from                { shift->_header('From',                @_); }
sub if_match            { shift->_header('If-Match',            @_); }
sub if_none_match       { shift->_header('If-None-Match',       @_); }
sub accept              { shift->_header('Accept',              @_); }
sub accept_charset      { shift->_header('Accept-Charset',      @_); }
sub accept_encoding     { shift->_header('Accept-Encoding',     @_); }
sub accept_language     { shift->_header('Accept-Language',     @_); }
sub referer             { shift->_header('Referer',             @_); }
sub user_agent          { shift->_header('User-Agent',          @_); }
sub mime_version        { shift->_header('MIME-Version',        @_); }
sub range               { shift->_header('Range',               @_); }
sub cookie              { shift->_header('Cookie',              @_); }
sub warning             { shift->_header('Warning',             @_); }
sub www_authenticate    { shift->_header('WWW-Authenticate',    @_); }
sub authorization       { shift->_header('Authorization',       @_); }
sub proxy_authenticate  { shift->_header('Proxy-Authenticate',  @_); }
sub proxy_authorization { shift->_header('Proxy-Authorization', @_); }
sub basic_authorization { shift->_basic_auth('Authorization',       @_); }
sub basic_proxy_auth    { shift->_basic_auth('Proxy-Authorization', @_); }

#---------------------------------------------
#   Base64 Encode function
#---------------------------------------------
eval {
    require MIME::Base64; MIME::Base64->import('encode_base64');
};
if($@){
    *encode_base64 = \&perl_encode_base64;
}

sub perl_encode_base64 ($;$) {
    my $str = shift;
    my($eol) = @_;
    $eol = "\n" unless defined $eol;
    my $res = "";
    pos($str) = 0;
    while ($str =~ /(.{1,45})/gso){
        $res .= substr(pack('u', $1), 1);
        chop($res);
    }
    $res =~ tr|` -_|AA-Za-z0-9+/|;
    if(my $pad = (3 - length($str) % 3) % 3){
        $res =~ s/.{$pad}$/'=' x $pad/e;
    }
    if(length($eol)){
        $res =~ s/(.{1,76})/$1$eol/g;
    }
    $res;
}

#--------------------------------------------
#   Debug option support
#--------------------------------------------
sub set_debug {
    my($self, $mode) = @_;
    $self->{debug} = $mode ? 1 : 0;
}

sub get_debug {
    my $self = shift;
    foreach (@{$self->{logs}}){
        s/$CRLF/\n/sgo;
    }
    unshift(@{$self->{logs}}, "\n");
    unshift(@{$self->{logs}}, join('', 'proxy server: ', $self->{proxy}, "\n")) if defined $self->{proxy};
    unshift(@{$self->{logs}}, join('', 'httpd server: ', $self->{host}, ':', $self->{port}, "\n"));
    unshift(@{$self->{logs}}, join('', 'status code : ', join(' ', @{$self->{status}}, "\n")));
    return @{$self->{logs}};
}

1;

__END__
