#=====================================================================
#   POP3.pm : Simple POP3 client module with MIME support.
#
#   Copyright(c) 2001, Nobuchika Oishi (BSC CONSULTING).
#
#   e-mail : bigstone@my.email.ne.jp
#   support: http://www.din.or.jp/~bigstone/cgilab/index.html
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND
#   ANY EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED.
#
#   ver 1.0.1 : updated last on 2001/09/24
#
#=====================================================================
package POP3;

use strict;
use integer;

use Carp;
use TCP;
use MIME;

use vars qw($VERSION $MAX_RECV_SIZE $UUDECODE_TEXT $CRLF_ANY);

#---------------------------------------------------------------------
$MAX_RECV_SIZE = 3 * 1024 * 1024;   # default value. (bytes)
$UUDECODE_TEXT = 1;                 # default value. {0|1|2}
#---------------------------------------------------------------------

$VERSION  = '1.0.1';
$CRLF_ANY = "\015?\012";

sub new {
    my $class = shift;
    my $self = {
        maxsize  => $MAX_RECV_SIZE,
        uudecode => $UUDECODE_TEXT,
        suffix   => [],
        logs     => [],
    };
    my($k, $v);
    while (($k, $v) = splice(@_, 0, 2)){
        $self->{lc($k)} = $v;
    }
    $self->{socket} = TCP->connect($self->{host}, $self->{port} || 110, $self->{timeout}) || return undef;
    bless $self, ref $class || $class;
    my $debug = $self->{debug}; $self->{debug} = 1;
    unless($self->_command()){
        $self->close();
        return undef;
    }
    if($self->{status} =~ /(<[^>]+>)/o){
        $self->{banner} = $1;
    }
    $self->{debug} = $debug;
    return $self;
}

sub login {
    my $self = shift;
    my($user, $pass) = @_;
    $self->{user} = $user;
    $self->user($user) and $self->pass($pass);
}

sub user {
    shift->_command('USER', $_[0]);
}

sub pass {
    my $self = shift;
    my $status = $self->_command('PASS', $_[0]) || return undef;
    $self->stat();
    return $status;
}

sub stat {
    my $self = shift;
    unless(defined $self->{count}){
        my $status = $self->_command('STAT') || return undef;
        my($count, $size) = $status =~ /(\d+)\D+(\d+)/o;
        $self->{count} = $count || 0;
        $self->{size}  = $size  || 0;
    }
    return wantarray ? ($self->{count}, $self->{size}) : $self->{count};
}

sub list {
    my $self = shift;
    return undef unless $self->{count};
    if(defined $self->{flag}){
        return $self->{list}->{$_[0]} if @_;
        return wantarray ? %{$self->{list}} : $self->{list};
    }
    $self->{list} = {} unless defined $self->{list};
    if(@_ && exists $self->{list}->{$_[0]}){
         return $self->{list}->{$_[0]};
    }
    my $status = $self->_command('LIST', @_) || return undef;
    if(@_){
        if($status =~ /(\d+)\D+(\d+)/o){
            $self->{list}->{$1} = $2;
            return $2;
        } else {
            return undef;
        }
    }
    my $response = $self->_response() || return undef;
    %{ $self->{list} } = map { /(\d+)\D+(\d+)/o } split(/\n/, $response);
    $self->{flag} = 1;
    return wantarray ? %{$self->{list}} : $self->{list};
}

sub head {
    my $self = shift;
    my($num, $ocode) = @_;
    $self->top($num, 0, $ocode);
}

sub top {
    my $self = shift;
    my($num, $lines, $ocode) = @_; $lines ||= 0;
    $self->_init_result();
    return undef unless $self->{count};
    $self->{msgsize} = $self->list($num) || return undef;
    $self->_command('TOP', $num, $lines) || return undef;
    my $mime = MIME->new(
        socket => $self->{socket}
    );
    $mime->decode_message(convert => $ocode);
    $self->{mime} = $mime;
    return wantarray ? ($mime->header()) : $self->{msgsize};
}

sub retr {
    my $self = shift;
    my($num, $attach, $ocode, $dele) = @_;
    $self->_init_result();
    return undef unless $self->{count};
    my $size = $self->list($num) || return undef;
    if($size > $self->{maxsize}){
        if($self->{debug}){
            Carp::croak(join(' ', "Mail size limit: $size bytes. max", $self->{maxsize}));
        }
        return undef;
    }
    $self->_command('RETR', $num) || return undef;
    my $mime = MIME->new(
        socket => $self->{socket}
    );
    $mime->decode_message(
        convert  => $ocode,
        attach   => $attach,
        suffix   => $self->{suffix},
        uudecode => $self->{uudecode},
    );
    $self->{mime} = $mime;
    if($dele =~ /^dele$/oi){
        $self->dele($num) || return undef;
    }
    $self->{msgsize} = $size;
}

sub retr_nodecode {
    my $self = shift;
    my($num, $fh, $dele) = @_;
    $self->_init_result();
    return undef unless $self->{count};
    if(defined $fh){
        return undef unless defined fileno($fh);
    }
    my $size = $self->list($num) || return undef;
    if($size > $self->{maxsize}){
        if($self->{debug}){
            Carp::croak(join(' ', "Mail size limit: $size bytes. max", $self->{maxsize}));
        }
        return undef;
    }
    $self->_command('RETR', $num) || return undef;
    my $sock = $self->{socket};
    my($mail, $line, $eof);
    $size = 0;
    while ($_ = $sock->read()){
        s/$CRLF_ANY$/\n/so;
        if(/^\n$/so){
            $line .= $_; next;
        }
        if(/^\.\n$/so){
            $line =~ s/\n+$/\n/so; $eof = 1;
        }
        $size += length($line);
        if(defined $fh){
            unless(print $fh $line){
                Carp::croak("Can't output to file: $!") if $self->{debug};
                return undef;
            }
        } else {
            $mail .= $line;
        }
        last if $eof;
        $line = $_;
    }
    if($dele =~ /^dele$/oi){
        $self->dele($num) || return undef;
    }
    return (defined $fh) ? $size : $mail;
}

sub uidl {
    my $self = shift;
    my $status = $self->_command('UIDL', $_[0]) || return undef;
    if($_[0]){
        $status =~ /\d+\s+([\x21-\x7e]+)/o;
        return $1;
    }
    my $response = $self->_response() || return undef;
    my($num, $uid);
    my %uidl = ();
    foreach (split(/\n/, $response)){
        ($num, $uid) = /^\s*(\d+)\s+([\x21-\x7e]+)/o;
        $uidl{$num} = $uid;
    }
    return %uidl;
}

sub last {
    my $status = shift->_command('LAST') || return undef;
    return ($status =~ /(\d+)/o) ? $1 : 0;
}

sub dele {
    my $self = shift;
    my $num  = shift;
    my $status = $self->_command('DELE', $num) || return undef;
    $self->{dele} = {} unless defined $self->{dele};
    $self->{dele}->{$num} = $num;
    return $status;
}

sub reset {
    my $self = shift;
    my $status = $self->_command('RSET') || return undef;
    delete $self->{dele};
    return $status;
}

sub noop { shift->_command('NOOP'); }

sub quit {
    my $self = shift;
    my $status = $self->_command('QUIT') || return undef;
    $self->close();
    return $status;
}

sub close {
    my $self = shift;
    if(my $sock = delete $self->{socket}){
        $sock->close();
    }
}

sub apop {
    my $self = shift;
    eval {
        require Digest::MD5; Digest::MD5->import('md5_hex');
    };
    if($@){
        Carp::croak("You need to install Digest-MD5 to use the APOP.") if $self->{debug};
        return undef;
    }
    my($user, $pass) = @_;
    my $banner = $self->{banner} || return undef;
    my $status = $self->_command('APOP', $user, md5_hex($banner, $pass)) || return undef;
    $self->stat();
    return $status;
}

sub banner {
    shift->{banner};
}

sub dlist {
    my $self = shift;
    return undef unless defined $self->{dele};
    return (sort {$a <=> $b} keys %{$self->{dele}});
}

sub header  { shift->{mime}->header(@_);  }
sub content { shift->{mime}->content(@_); }
sub charset { shift->{mime}->charset(@_); }
sub attach  { shift->{mime}->attach(@_);  }

sub files {
    my $self = shift;
    scalar keys(%{$self->{attach}}) || 0;
}

sub size {
    shift->{msgsize};
}

sub status {
    shift->{status};
}

sub is_success {
    shift->{status} =~ /^\+OK/o;
}

sub clear {
    my $self = shift;
    delete $self->{count};
    delete $self->{size};
    delete $self->{list};
    delete $self->{dele};
    $self->_init_result();
}

sub _init_result {
    my $self = shift;
    $self->{header}  = [];
    $self->{content} = {};
    $self->{charset} = {};
    $self->{attach}  = {};
    $self->{msgsize} = 0;
    delete $self->{mime};
}

sub _command {
    my $self = shift;
    my $cmd  = shift;
    my $sock = $self->{socket} || return undef;
    $self->{status} = undef;
    if($cmd){
        $cmd = join(' ', $cmd, @_) if @_;
        push(@{$self->{logs}}, "$cmd\n") if $self->{debug};
        unless($sock->write("$cmd\n")){
            Carp::croak("POP3 command send error: $cmd") if $self->{debug};
            return undef;
        }
    }
    my $status = $sock->read() || return undef;
    $status =~ s/[\r\n]//sgo;
    $self->{status} = $status;
    push(@{$self->{logs}}, "$status\n") if $self->{debug};
    return ($status =~ /^\+OK/o) ? $status : undef;
}

sub _response {
    my $self = shift;
    my $sock = $self->{socket} || return undef;
    my $res  = $sock->read_until_dot() || return undef;
    $res =~ s/$CRLF_ANY/\n/sgo;
    return $res;
}

#----------------------------------------------
#   Debug option support
#----------------------------------------------
sub debug {
    my $self = shift;
    if(@_){
        $self->{debug} = shift;
        if($self->{debug} == -1){
            $self->{logs} = [];
        }
    } else {
        return wantarray ? @{$self->{logs}} : $self->{logs};
    }
}

1;

__END__
